using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using AssetBundles;
using Battlehub.RTCommon;
using SimpleJSON;
using Stopper;
using UnityEngine;
using UnityEngine.UI;
using MVR.FileManagementSecure;
using Battlehub.RTSaveLoad;
using Battlehub.RTCommon;
using UnityStandardAssets.ImageEffects;
/// <summary>

/// </summary>

namespace Stopper {
	public class CUAControllers : MVRScript {
		string savePath = "Saves\\CUAController";
		string fileExt = "cuacontroller";

		// CUA Loader
		CustomUnityAssetLoader loader;

		// Prefabricated UI for joint controllers. Not sure how to get ahold of this
		// other than copying from somewhere
		GameObject controllerUIPrefab;

		// A sample controller for retrieving controller meshes. Again, not sure
		// how to find this other than copying from elsewhere
		FreeControllerV3 sampleController;

		// GameObjects to organize all of our stuff
		GameObject unityAssetControls;
		GameObject controlUI;
		GameObject controllers;

		Atom controlAtom;
		string controlAtomUid;
		Transform cuaRoot;
		Transform root;

		List<FreeControllerV3> createdControllers = new List<FreeControllerV3>();
		List<ConfigurableJoint> controllerJoints = new List<ConfigurableJoint>();

		List<ConfigurableJoint> createdJoints = new List<ConfigurableJoint>();
		List<Rigidbody> rigRigidbodies = new List<Rigidbody>();
		List<Rigidbody> controlRigidbodies = new List<Rigidbody>();

		List<UnityEngine.Object> createdObjects = new List<UnityEngine.Object>();

		// UI
		string selectedRootPath;

		JSONStorableBool createRigidBodies;
		JSONStorableBool createJoints;
		JSONStorableBool loadOnStartup;

		bool loaded = false;
		bool isCUA = false;

		JSONClass savedJson;

		float defaultRbMass = 1f;
		float defaultRbDrag = 1f;
		float defaultJointDriveSpring = 100f;
		float defaultJointDriveDamper = 0f;
		float defaultJointDriveMaxForce = 100f;

		Dictionary<string, FreeControllerV3> rigidBodyControllers = new Dictionary<string, FreeControllerV3>();

		public override void Init() {
			CreateControlAtom();

			rigidBodyControllers = new Dictionary<string, FreeControllerV3>();
			loader = containingAtom.reParentObject.GetComponentInChildren<CustomUnityAssetLoader>(true);
			if (loader == null) {
				isCUA = false;
			} else {
				isCUA = true;
			}

			GameObject windowDisplayControl = GameObject.Find("/SceneAtoms/WindowCamera/reParentObject/DisplayControl");
			if (windowDisplayControl == null) {
				SuperController.LogError("Unable to find window display control.");
				return;
			}

			sampleController = windowDisplayControl.GetComponent<FreeControllerV3>();

			if (controllerUIPrefab == null) {
				FindSampleControllerUI();
			}

			if (controllerUIPrefab == null) {
				var personAtom = LoadAssetBundleSync("a_per", "Person");
				FindSampleControllerUI();
			}

			if (controllerUIPrefab == null || sampleController == null) {
				SuperController.LogError("Unable to load required templates. Fix this by adding a Person to the scene");
				return;
			}

			JSONStorableStringChooser rootChooser = new JSONStorableStringChooser(
				"Root Node",
				GetRootCandidates(),
				null,
				"Root Node",
				(string path) => {
					selectedRootPath = path;
				});
			CreateScrollablePopup(rootChooser, false);
			RegisterStringChooser(rootChooser);


			createRigidBodies = new JSONStorableBool("Create Rigidbodies", true);
			createJoints = new JSONStorableBool("Create Joints", true);

			var createRigidBodiesToggle = CreateToggle(createRigidBodies, false);
			RegisterBool(createRigidBodies);

			var createJointsToggle = CreateToggle(createJoints, false);
			RegisterBool(createJoints);

			loadOnStartup = new JSONStorableBool("LoadOnStartup", false);
			RegisterBool(loadOnStartup);

			UIDynamicButton goButton = CreateButton("Create Controllers", false);
			goButton.button.onClick.AddListener(() => {
				AddControllers(false);
			});

			// var loadPresetUI = CreateButton("Load Preset");
			// loadPresetUI.button.onClick.AddListener(() => {
			// 	FileManagerSecure.CreateDirectory(savePath);
			// 	var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
			// 	SuperController.singleton.GetMediaPathDialog(LoadPreset, fileExt, savePath, false, true, false, null, false, shortcuts);
			// });

			// var savePresetUI = CreateButton("Save Preset");
			// savePresetUI.button.onClick.AddListener(() => {
			// 	FileManagerSecure.CreateDirectory(savePath);
			// 	var fileBrowserUI = SuperController.singleton.fileBrowserUI;
			// 	fileBrowserUI.SetTitle("Save Preset");
			// 	fileBrowserUI.fileRemovePrefix = null;
			// 	fileBrowserUI.hideExtension = false;
			// 	fileBrowserUI.keepOpen = false;
			// 	fileBrowserUI.fileFormat = fileExt;
			// 	fileBrowserUI.defaultPath = savePath;
			// 	fileBrowserUI.showDirs = true;
			// 	fileBrowserUI.shortCuts = null;
			// 	fileBrowserUI.browseVarFilesAsDirectories = false;
			// 	fileBrowserUI.SetTextEntry(true);
			// 	fileBrowserUI.Show(SavePreset);
			// 	fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + fileExt;
			// 	fileBrowserUI.ActivateFileNameField();
			// });

			AddSlider("Rigidbody Drag", defaultRbDrag, 0f, 2f, (float val) => {
				foreach (var rb in rigRigidbodies) {
					rb.drag = val;
				}
			}, true, true);

			AddSlider("Rigidbody Mass", defaultRbMass, 0f, 2f, (float val) => {
				foreach (var rb in rigRigidbodies) {
					rb.mass = val;
				}
			}, true, true);

			AddSlider("Joint Spring", defaultJointDriveSpring, 0f, 1000f, (float val) => {
				foreach (var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.positionSpring = val;
					joint.slerpDrive = drive;
				}
			}, true, true);

			AddSlider("Joint Damper", defaultJointDriveDamper, 0f, 10f, (float val) => {
				foreach (var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.positionDamper = val;
					joint.slerpDrive = drive;
				}
			}, true, true);

			AddSlider("Joint Max Force", defaultJointDriveMaxForce, 0f, 1000f, (float val) => {
				foreach (var joint in createdJoints) {
					var drive = joint.slerpDrive;
					drive.maximumForce = val;
					joint.slerpDrive = drive;
				}
			}, true, true);
		}

		private List<String> GetRootCandidates() {
			List<string> candidates = new List<string>();

			Queue<Transform> frontier = new Queue<Transform>();

			var asset = containingAtom.GetStorableByID("asset");
			if (asset != null) {
				cuaRoot = asset.transform;
			}
			if (cuaRoot == null) {
				cuaRoot = containingAtom.reParentObject.Find("object/rescaleObject");
			}
			if (cuaRoot == null) {
				cuaRoot = containingAtom.reParentObject;
			}

			frontier.Enqueue(cuaRoot);

			Transform node;
			while (frontier.Count > 0) {
				node = frontier.Dequeue();
				if (frontier == null)
					break;
				for (int i = 0; i < node.childCount; i++) {
					var child = node.GetChild(i);
					if (child.childCount == 0) {
						continue;
					}
					if (candidates.Count < 30) {
						if(child.name != unityAssetControls.name ) {
							candidates.Add(GetPathToRoot(child, cuaRoot));
						}
						frontier.Enqueue(child);
					} else {
						return candidates;
					}
				}
			}

			return candidates;
		}

		private string GetPathToRoot(Transform t, Transform root) {
			if (t.parent == null || t.parent == root) {
				return t.name;
			} else {
				return GetPathToRoot(t.parent, root) + "/" + t.name;
			}
		}

		private GameObject LoadAssetBundleSync(string assetBundleName, string assetName, bool deactivate = true) {
			GameObject asset = null;

			string error;
			var loadedAssetBundle = AssetBundleManager.GetLoadedAssetBundle(assetBundleName, out error);

			if (loadedAssetBundle == null) {
				AssetBundleLoader assetLoader = new AssetBundleLoader();
				assetLoader.Load(assetBundleName, (bundleName, assetBundle) => {
					var loadedAsset = assetBundle.LoadAsset(assetName, typeof(GameObject)) as GameObject;

					if (deactivate) {
						loadedAsset.SetActive(false);
					}

					asset = UnityEngine.Object.Instantiate<GameObject>(loadedAsset);

					if (deactivate) {
						loadedAsset.SetActive(true);
					}

					assetBundle.Unload(true);
				});
			} else {
				var loadedAsset = loadedAssetBundle.m_AssetBundle.LoadAsset(assetName, typeof(GameObject)) as GameObject;
				if (deactivate) {
					loadedAsset.SetActive(false);
				}
				asset = UnityEngine.Object.Instantiate<GameObject>(loadedAsset);
				if (deactivate) {
					loadedAsset.SetActive(true);
				}
			}

			return asset;
		}

		private void CreateControlAtom() {
			if(controlAtom == null) {
				controlAtom = this.containingAtom;

				unityAssetControls = CreateChildObject(controlAtom.reParentObject.gameObject, "CustomUnityAssetControls");
				controlUI = CreateChildObject(controlAtom.gameObject, "CustomUnitAssetControlsUI");
				controllers = CreateChildObject(unityAssetControls, "Controllers");
			}
		}

		/*
			If you are reading this, then you have my condolences. 

			An Atom's controllers and rigidbodies are saved in its internal lists (populated during Init), 
			as well as some structures within the SuperController (populated in AddAtom). 

			Controllers won't work unless both of these are updated. 
			
			Manually calling Init via SendMessage can populate an Atom's internal state, but the SuperController
			state requires it to be re-added. An Atom can't be added twice, so first we have to remove it without
			deleting it. Luckily VAM's atom pooling logic helps with that, but it still resets some internal state
			that we'll need to try to preserve. 

			This method attempts to navigate removing and re-adding an Atom while maintaining all of the state.
		*/
		private void ReAddAtom(bool fromRestore = false, bool teardown = false) {
			// Set the atom to be poolable, even if it isn't typically.
			// This will prevent VAM from destroying it on removal.
			bool isPoolable = controlAtom.isPoolable;
			controlAtom.isPoolable = true;

			// Removing an atom deselects it
			bool reSelect = SuperController.singleton.GetSelectedAtom() == controlAtom;

			// Removing an atom clears its parent status, so we'll need to restore that.
			Atom parentAtom = controlAtom.parentAtom;
			


			string uid = controlAtom.uid;

			// During teardown, remove the created controllers from storables list so that we don't
			// try to preserve them later
			if(teardown) {
				foreach (FreeControllerV3 controller in createdControllers) {
					controlAtom.UnregisterAdditionalStorable(controller);
				}
			}


			// Take a snapshot of the current state of the atom. This should more or less 
			// be the same as the current lastRestoreData, which we'll lose during Remove
			JSONArray jsonArray = new JSONArray();
			controlAtom.Store(jsonArray, true, true);
			JSONClass restoreJson = jsonArray[0].AsObject;

			// RemoveAtom clears its storables, so we're going to unregister them
			// and re-add later
			List<JSONStorable> storables = new List<JSONStorable>();
			foreach (var storableId in controlAtom.GetStorableIDs()) {
				JSONStorable storable = controlAtom.GetStorableByID(storableId);
				storables.Add(storable);
				controlAtom.UnregisterAdditionalStorable(storable);
				storable.containingAtom = controlAtom;
			}


			// Do the remove. 
			SuperController.singleton.RemoveAtom(controlAtom);
			
			// During teardown, we'll need to wait until after RemoveAtom but before Init to delete these
			// This allows Remove to de-register them from the SuperController, and when Init runs it excludes
			// them from the Atom's internal state.
			if(teardown) {
				foreach (var o in createdControllers) {
					DestroyImmediate(o);
				}

				// Destroy rigidbodies/joints
				foreach (var o in controllerJoints) {
					DestroyImmediate(o);
				}

				foreach (var o in createdJoints) {
					DestroyImmediate(o);
				}

				foreach (var o in rigRigidbodies) {
					DestroyImmediate(o);
				}

				foreach (var o in controlRigidbodies) {
					DestroyImmediate(o);
				}
			}


			controlAtom.gameObject.SetActive(true);
			controlAtom.inPool = false;

			// Restore the lastRestoreData and re-add the preserved storables
			controlAtom.Restore(restoreJson, true, true, true, null, false, false, false, false);
			foreach (var storable in storables) {
				controlAtom.RegisterAdditionalStorable(storable);
			}
			
			// Re-init the atom to update its state
			controlAtom.SendMessage("Init");

			// Add it back to the pool
			SuperController.singleton.AddAtom(controlAtom, uid, false, reSelect, false, false);

			// Restore parent
			if(parentAtom != null) {
				controlAtom.SetParentAtom(parentAtom);
			}

			// If the atom wasn't poolable before, we want to make sure it doesn't get 
			// pooled in the future
			controlAtom.isPoolable = isPoolable;
		}

		private void AddControllers(bool fromRestore = true) {
			
			if (loaded) {
				SuperController.LogMessage("Already loaded. Reload plugin if needed.");
				return;
			}

			if (selectedRootPath == null) {
				SuperController.LogMessage("No root node selected.");
				return;
			}

			root = cuaRoot.Find(selectedRootPath);

			if (root == null) {
				SuperController.LogMessage($"Could not find root node at path {selectedRootPath}");
				return;
			}

			ProcessRigging(root, 0);

			foreach (Rigidbody rb in root.GetComponentsInChildren<Rigidbody>(true)) {
				AddController(rb.name);
			}

			if (rigidBodyControllers.ContainsKey(root.name)) {
				rigidBodyControllers[root.name].currentPositionState = FreeControllerV3.PositionState.On;
				rigidBodyControllers[root.name].currentRotationState = FreeControllerV3.RotationState.On;
			}

			// We want the controllers under the rescaleObject, but it seems to get cleared
			// at some point, so we're waiting until now to do it
			var rescaleObject =  controlAtom.reParentObject.Find("object/rescaleObject");
			if(rescaleObject != null) {
				unityAssetControls.transform.parent = rescaleObject;
			}


			// Set the controller state from the lastRestoredData that was populated during scene load
			// We're waiting until now to do that since we want to make sure other controllers were loaded
			if(fromRestore) {
				foreach (var controller in createdControllers) {
					controlAtom.RestoreFromLast(controller);
				}
			}

			// Remove and re-add the atom to populate SuperController maps
			ReAddAtom(fromRestore);

			foreach (Rigidbody rb in root.GetComponentsInChildren<Rigidbody>(true)) {
				LinkRBToController(rb, fromRestore);
			}
			
			foreach(var joint in controllerJoints){
				var rb = joint.connectedBody;
				joint.connectedBody = null;
				joint.connectedBody = rb;
			}
			


			
			if(fromRestore) {
				StartCoroutine("RestoreAllControllers");
			}

			loaded = true;
			loadOnStartup.val = true;
		}

		private IEnumerator RestoreAllControllers() {
			yield return new WaitForFixedUpdate();
			//yield return new WaitForSeconds(10f);
			foreach (var atom in SuperController.singleton.GetAtoms()) {
				if (atom == controlAtom) continue;
				foreach (var controller in atom.freeControllers) {
					atom.RestoreFromLast(controller);
				}
			}
			yield break;
		}

		private void FindSampleControllerUI() {
			foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]) {
				if (go.name == "UIFreeControllerJointNoDrive") {
					controllerUIPrefab = go;
					break;
				}
			}
		}

		private Rigidbody ProcessRigging(Transform parent, int depth) {
			Rigidbody parentRB = parent.gameObject.GetComponent<Rigidbody>();
			if (parentRB == null && createRigidBodies.val) {
				parentRB = CreateComponent(parent.gameObject, typeof(Rigidbody)) as Rigidbody;
				parentRB.isKinematic = false;
				parentRB.useGravity = false;
				parentRB.detectCollisions = true;
				parentRB.drag = defaultRbDrag;
				parentRB.mass = defaultRbMass;
				rigRigidbodies.Add(parentRB);
			}

			try {
				//	SuperController.LogMessage($"Processing: {parent}");
				for (int i = 0; i < parent.childCount; i++) {
					Transform child = parent.GetChild(i);
					//		SuperController.LogMessage($"Child: {child}");

					Rigidbody childRB = child.GetComponent<Rigidbody>();
					if (childRB == null) {
						if (createRigidBodies.val) {
							childRB = CreateComponent(child.gameObject, typeof(Rigidbody)) as Rigidbody;
							childRB.isKinematic = false;
							childRB.useGravity = false;
							childRB.detectCollisions = true;
							childRB.drag = defaultRbDrag;
							childRB.mass = defaultRbMass;
							rigRigidbodies.Add(childRB);
						} else {
							continue;
						}
					}

					if (parentRB != null && createJoints.val) {

						ConfigurableJoint childJoint = CreateComponent(child.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
						childJoint.connectedBody = parentRB;
						createdJoints.Add(childJoint);

						childJoint.xMotion = ConfigurableJointMotion.Locked;
						childJoint.yMotion = ConfigurableJointMotion.Locked;
						childJoint.zMotion = ConfigurableJointMotion.Locked;

						childJoint.angularXMotion = ConfigurableJointMotion.Limited;
						childJoint.angularYMotion = ConfigurableJointMotion.Limited;
						childJoint.angularZMotion = ConfigurableJointMotion.Limited;

						childJoint.rotationDriveMode = RotationDriveMode.Slerp;
						JointDrive slerpDrive = new JointDrive();
						slerpDrive.positionSpring = defaultJointDriveSpring;
						slerpDrive.positionDamper = defaultJointDriveDamper;
						slerpDrive.maximumForce = defaultJointDriveMaxForce;
						childJoint.slerpDrive = slerpDrive;

						SoftJointLimit xLowJointLimit = new SoftJointLimit();
						xLowJointLimit.limit = -180f;
						SoftJointLimit jointLimit = new SoftJointLimit();
						jointLimit.limit = 180f;
						childJoint.lowAngularXLimit = xLowJointLimit;
						childJoint.highAngularXLimit = jointLimit;
						childJoint.angularYLimit = jointLimit;
						childJoint.angularZLimit = jointLimit;

					}

					ProcessRigging(child, depth + 1);
				}

			} catch (Exception ex) {
				SuperController.LogError($"Failed to init UnityObjectController: {ex}");
			}
			return parentRB;
		}

		private bool AddController(string rigidBodyName) {
			try {
				if (rigidBodyControllers.ContainsKey(rigidBodyName)) {
					// already exists
					return false;
				}

				GameObject rbControl = CreateChildObject(controllers, $"{rigidBodyName}Control");
				rbControl.SetActive(false);
				rbControl.layer = 8;

				Rigidbody controlRB = CreateComponent(rbControl, typeof(Rigidbody)) as Rigidbody;
				controlRB.isKinematic = true;
				controlRB.useGravity = false;
				controlRB.detectCollisions = true;
				controlRigidbodies.Add(controlRB);

				/*	
				ConfigurableJoint controlJoint = CreateComponent(rb.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
				controlJoint.connectedBody = controlRB;
				controlJoint.autoConfigureConnectedAnchor = false;
				controlJoint.anchor = Vector3.zero;
				controlJoint.connectedAnchor = Vector3.zero;
				
				if( !controllerJoints.Contains(controlJoint)){
					controllerJoints.Add(controlJoint);
				}
			*/
				FreeControllerV3 controller = CreateComponent(rbControl, typeof(FreeControllerV3)) as FreeControllerV3;
				controller.containingAtom = controlAtom;
				//controller.followWhenOff = rb.transform;
				controller.useGravityOnRBWhenOff = false;
				controller.meshScale = 0.5f;
				controller.selectedScale = 0.1f;
				controller.enableSelectRoot = true;
				controller.enabled = true;
				controller.hidden = false;
				controller.startingPositionState = FreeControllerV3.PositionState.Off;
				controller.startingRotationState = FreeControllerV3.RotationState.Off;

				if (!createdControllers.Contains(controller)) {
					createdControllers.Add(controller);
				}
				rigidBodyControllers[rigidBodyName] = controller;

				// Copy meshes from the sample controller
				controller.material = sampleController.material;
				if (controller.onPositionMesh == null) {
					controller.onPositionMesh = sampleController.onPositionMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.onPositionMesh, controller.transform);
					controller.offPositionMesh = sampleController.offPositionMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.offPositionMesh, controller.transform);
					controller.followingPositionMesh = sampleController.followingPositionMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.followingPositionMesh, controller.transform);
					controller.holdPositionMesh = sampleController.holdPositionMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.holdPositionMesh, controller.transform);
					controller.lockPositionMesh = sampleController.lockPositionMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.lockPositionMesh, controller.transform);
					controller.onRotationMesh = sampleController.onRotationMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.onRotationMesh, controller.transform);
					controller.offRotationMesh = sampleController.offRotationMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.offRotationMesh, controller.transform);
					controller.followingRotationMesh = sampleController.followingRotationMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.followingRotationMesh, controller.transform);
					controller.holdRotationMesh = sampleController.holdRotationMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.holdRotationMesh, controller.transform);
					controller.lockRotationMesh = sampleController.lockRotationMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.lockRotationMesh, controller.transform);
					controller.lookAtRotationMesh = sampleController.lookAtRotationMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.lookAtRotationMesh, controller.transform);
					controller.moveModeOverlayMesh = sampleController.moveModeOverlayMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.moveModeOverlayMesh, controller.transform);
					controller.rotateModeOverlayMesh = sampleController.rotateModeOverlayMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.rotateModeOverlayMesh, controller.transform);
					controller.deselectedMesh = sampleController.deselectedMesh == null ? null : UnityEngine.Object.Instantiate<Mesh>(sampleController.deselectedMesh, controller.transform);
				}

				SphereCollider collider = CreateComponent(rbControl, typeof(SphereCollider)) as SphereCollider;
				collider.enabled = true;
				collider.radius = 0.1f;

				// Controller Link. Not sure what this is used for.
				GameObject controlLink = CreateChildObject(rbControl, rigidBodyName + "Link");
				controlLink.layer = 8;

				SphereCollider linkCollider = CreateComponent(controlLink, typeof(SphereCollider)) as SphereCollider;
				linkCollider.enabled = true;
				linkCollider.radius = 0.01f;

				Rigidbody linkRB = CreateComponent(controlLink, typeof(Rigidbody)) as Rigidbody;
				linkRB.isKinematic = true;
				linkRB.useGravity = false;
				linkRB.detectCollisions = false;
				controlRigidbodies.Add(linkRB);

				FreeControllerV3Link freeControllerLink = CreateComponent(controlLink, typeof(FreeControllerV3Link)) as FreeControllerV3Link;
				freeControllerLink.linkedController = controller;


				// UI
				GameObject uiPlaceholder = CreateChildObject(controlUI, $"{rigidBodyName}UIPlaceholder");
				uiPlaceholder.SetActive(false);

				MoveAndRotateAsHUDAnchor moveAsHud = CreateComponent(uiPlaceholder, typeof(MoveAndRotateAsHUDAnchor)) as MoveAndRotateAsHUDAnchor;
				moveAsHud.scaleAsAnchor = true;
				moveAsHud.anchorNum = HUDAnchor.AnchorNum.Two;

				UIConnectorMaster connectorMaster = CreateComponent(uiPlaceholder, typeof(UIConnectorMaster)) as UIConnectorMaster;
				connectorMaster.UIPrefab = controllerUIPrefab.transform;

				UIConnector connector = CreateComponent(uiPlaceholder, typeof(UIConnector)) as UIConnector;
				connector.receiver = controller;
				connector.receiverTransform = rbControl.transform;

				controller.UITransforms = new Transform[] { uiPlaceholder.transform };
				controller.UITransformsPlayMode = new Transform[] { };

				controlRB.gameObject.SetActive(true);
				return true;
			} catch (Exception e) {
				SuperController.LogMessage($"Failed to add controller for {rigidBodyName} {e}");
				return false;
			}
		}

		public void LinkRBToController(Rigidbody rb, bool rigidbodiesFollowControllers = false) {
			if (!rigidBodyControllers.ContainsKey(rb.name)) {
				SuperController.LogError($"Controller does not exist for rigidbody {rb.name}");
				return;
			}

			FreeControllerV3 controller = rigidBodyControllers[rb.name];
			Rigidbody controlRB = controller.gameObject.GetComponent<Rigidbody>();
			//SuperController.LogMessage($"Linking ${rb} to {controlRB}");
			if (controller.followWhenOff != null) {
				SuperController.LogMessage($"Controller {rb.name} already linked");
				return;
			}
			var pos = controlRB.transform.position;
			var rot = controlRB.transform.rotation;

			controlRB.transform.position = rb.transform.position;
			controlRB.transform.rotation = rb.transform.rotation;

			ConfigurableJoint controlJoint = CreateComponent(rb.gameObject, typeof(ConfigurableJoint)) as ConfigurableJoint;
			controlJoint.connectedBody = controlRB;
			controlJoint.autoConfigureConnectedAnchor = false;
			controlJoint.anchor = Vector3.zero;
			controlJoint.connectedAnchor = Vector3.zero;

			if (!controllerJoints.Contains(controlJoint)) {
				controllerJoints.Add(controlJoint);
			}

			controller.followWhenOffRB = rb;

			controller.gameObject.SetActive(true);
			
			if(rigidbodiesFollowControllers) {
				rb.transform.position = pos;
				rb.transform.rotation = rot;
				controller.transform.position = pos;
				controller.transform.rotation = rot;
			}
		}

		public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = true, bool rightSide = false) {
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			uiSlider.rangeAdjustEnabled = rangeAdjust;
			uiSlider.valueFormat = "F3";
			RegisterFloat(storableFloat);

			if (onChange != null) {
				storableFloat.setCallbackFunction = onChange;
			}

			return storableFloat;
		}

		private IEnumerator WaitForRootLoaded() {
			root = cuaRoot.Find(selectedRootPath);
			while (root == null) {
				yield return null;
				root = cuaRoot.Find(selectedRootPath);
			}
			while(SuperController.singleton.isLoading){
				yield return null;
			}

		//	yield return new WaitForSeconds(1f);
			AddControllers();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false) {
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);

			JSONArray rigidbodyNames = new JSONArray();
			foreach (var entry in rigidBodyControllers) {
				rigidbodyNames.Add(entry.Key);
			}
			jc["rigidbodyNames"] = rigidbodyNames;
			return jc;
		}

		public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true) {
			CreateControlAtom();

			if (jc["rigidbodyNames"] != null) {
				foreach (var rbJson in jc["rigidbodyNames"].AsArray) {
					AddController((rbJson as JSONNode).Value);
				}

				var reparentParent = controlAtom.reParentObject.parent;
				controlAtom.reParentObject.parent = controlAtom.transform;
				controlAtom.SendMessage("Init");
				controlAtom.reParentObject.parent = reparentParent;
			}

			base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
			savedJson = jc;

			StartCoroutine("WaitForRootLoaded");
		}

		public override void PostRestore() {

		}

		private void LoadPreset(string path) {
			if (string.IsNullOrEmpty(path))
				return;

			var presetJson = (JSONClass)LoadJSON(path);

			RestoreFromJSON(presetJson, true, true, null, false);
		}

		private void SavePreset(string path) {
			SuperController.singleton.fileBrowserUI.fileFormat = null;
			if (string.IsNullOrEmpty(path)) {
				return;
			}

			if (!path.ToLower().EndsWith($".{fileExt}")) {
				path += $".{fileExt}";
			}

			SaveJSON(GetJSON(true, true, true), path);
		}

		private GameObject FindGameObject(GameObject parent, string name) {
			Transform childTransform = parent.transform.Find(name);

			if (childTransform == null) {
				return null;
			} else {
				return childTransform.gameObject;
			}
		}

		private GameObject CreateChildObject(GameObject parent, string name) {
			GameObject gameObject = new GameObject(name);
			gameObject.transform.parent = parent.transform;
			createdObjects.Add(gameObject);

			return gameObject;
		}

		private Component CreateComponent(GameObject parent, System.Type type) {
			Component component = parent.AddComponent(type) as Component;
			createdObjects.Add(component);
			return component;
		}

		public void OnDestroy() {
			if(loaded) {
				ReAddAtom(false, true);
				
				controlAtom.UnregisterAdditionalStorable(this);
			}

			// Destroy the controller container
			foreach(var o in createdObjects){
				Destroy(o);
			}
		}
	}
}